# -*- coding: utf-8 -*-

from utilities.simulink_parser import *
import utilities.simulink_globals
import simulink_port
from utilities.simulink_utilities import *

# Main structure for a Simulink block: identifier, name, type, parameter(s), list of input and output ports and parent
class SimulinkBlock:
	fcar = [" ", "/", "(", ")", "[", "]", "\n", "&","-", ","]
	
	def __init__(self, xml = None, size_in = None, size_out = None, parent = None):

		# xml content of the block
		self.xml = xml
		
		# block identifier
		self.sid = None

		# block name
		self.name = None

		# block parameters
		self.params = []
		
		# block type as specified by Simulink
		self.type = None
		
		# block type: basic or composed
		self.ctype = "Basic"

		# block parent
		self.parent = parent

		self.del_inport = []
		self.skip = False
		self.basic = True
		
		# Instantiate block variables from the xml content
		if xml:
			
			if "SID" in xml.attrib.keys():
				self.sid = xml.attrib['SID']
			else:
				self.sid = 0
			if "Name" in xml.attrib.keys():
				aux = xml.attrib['Name']
				for c in SimulinkBlock.fcar:
					aux = aux.replace(c, "")
				self.name = aux.rstrip("_")
			if "BlockType" in xml.attrib.keys():
				self.type = self.xml.attrib['BlockType']
			ports = get_property(xml, 'Ports')
			size_in = 0
			size_out = 0
			if ports:
				ports = eval(ports)
				if ports:
					size_in = ports[0]
					if len(ports) > 1:
						size_out = ports[1]

		# Instantiate block inports and outports 
		self.create_ports(size_in, size_out)
		
		# Variable for the -nfb translation
		self.display_blocks = None

	
	# Create the requested number of input and output ports
	# if any of the parametres is None, then the corresponding port variable is not changed.
	def create_ports(self, size_in, size_out):
		# list of input ports
		if size_in != None:
			self.inport = [Port(self.sid, i, "in", 0) for i in range(0, size_in)]
		# list of output ports
		if size_out != None:
			self.outport = [Port(self.sid, i, "out", 0) for i in range(0, size_out)]

	# Split ports having arity more than one into multiple ports of arity one (both for inputs and outputs)
	def split_ports(self):
		def split(ports):
			split_ports = []
			for p in ports:
				split_ports.extend(p.split(len(split_ports)))
			return split_ports
		self.inport = split(self.inport)
		self.outport = split(self.outport)

	# Add state ports for blocks with state variables
	def add_state_ports(self):
		pass 

	# Set the initial value for blocks with state variables		
	def set_init(self):
		pass
		
	# Create new input ports up to an index and replace the index port with the corresponding value connected port
	# needed for enabled and triggered systems
	def set_in(self, index, value):
		n = len(self.inport)
		if index >= n:
			self.inport += [Port(self.sid, i, "in", 0) for i in range(n, index + 1)]
		self.inport[index] = value            

	# Create new output ports up to an index and replace the index port with the corresponding value connected port
	# needed for enabled and triggered systems
	def set_out(self, index, value):
		n = len(self.outport)
		if index >= n:
			self.outport += [Port(self.sid, i, "out", 0) for i in range(n, index + 1)]
		self.outport[index] = value

	# Add variable names to ports: a, b, ..., aa, ba, ca, ...
	def set_variables(self, var):
		for p in self.outport:
			p.set_var_name(var.get_var(p))
		for p in self.inport:
			if p.get_var_name() == None:
				p.set_var_name(var.get_var(p))
		
	# Update the arities of output ports and the corresponding state input port
	def update_arities(self):
		flag = 0
		for p in self.outport:
			p.set_arity(1)
		return flag

	# Split block with multiple outputs into multiple blocks with one output
	def split_block(self):
		return [self]

	# Obtain blocks on which the CPTS depends on
	def get_dependent_blocks(self):
		return []
		
	def get_id(self):
		return self.sid

	def get_name(self):
		return self.name

	def get_qualified_name(self):
		if self.parent == None:
			return self.get_name() if self.get_name() else 'Model' #self.get_name()
		else:
			return self.parent.get_qualified_name() + "_" + self.get_name()	

	def set_name(self, name):
		self.name = name
	
	def get_params(self):
		if len(self.params) == 0:
			return ""
		aux = " ".join(x for x in self.params)
		return aux
	
	def get_params_list(self):
		return self.params
	
	def set_params(self, p):
		if isinstance(p,list):
			for x in p:
				if x not in self.params and x != "":
					self.params.append(x)
		else:
			if p not in self.params and p != "":
				self.params.append(p)
		self.params.sort()

	# Generate parameters for Constant blocks in order to handle the "Additional types" error
	def update_const_params(self, opt_const, opt_not_generic):
		pass

	# Function not used anymore -- to remove
	'''
	def update_generic_params(self):
		pass
	'''
	
	# Update the values of symbolic constants, i.e., Constant c1 where c1 is defined in a text annotation
	def update_const_vals(self, cvals):
		pass

	def get_type(self):
		return self.type

	def get_ctype(self):
		return self.ctype
		
	def get_inport_arity(self, i):
		return self.inport[i].get_arity()
	   
	def set_inport_arity(self, i, value):
		self.inport[i].set_arity(value)

	def get_outport_arity(self, i):
		return self.outport[i].get_arity()

	def set_outport_arity(self, i, value):
		self.outport[i].set_arity(value)
		
	def get_nr_blocks(self):
		return 1
		
	def get_rename_base(self):
		return self

	def is_rename(self):
		return False

	def is_skip(self):
		return self.skip

	def is_basic(self):
		return self.basic
		
	# Generate the Isabelle expression either as a standalone component or used in another expression
	def isabelle(self, opt_fun, opt_generic, flag):
		# flag is true when the block is to be defined
		# flag is false when the block is used in a CPT
		
		# If flag is false used the block's qualified name
		if not flag:
			return iBasic(self.get_qualified_name())
		
		# Otherwise obtain the block's predicate transformer	
		return iBasic(self.get_type())

	# Generate the parameters of the Isabelle definition		
	def get_def_params(self, opt_fun, opt_generic):
		return self.get_params()
		
	# Generate the Isabelle definition of the component	
	def get_definition(self, opt_fun, opt_generic):
		return 'definition [%s]: "%s %s = %s"' % ('fblock' if self.get_type() == 'Fcn' else 'bblock', 
			self.get_qualified_name(), self.get_def_params(opt_fun, opt_generic), self.isabelle(opt_fun, opt_generic, True))
		
	def __repr__(self):
		return '{Block} %s, %s, %s, %s\n' % (self.get_id(), self.get_type(), self.inport, self.outport)

		
# Structure for a block resulting in Skip
class Skip(SimulinkBlock):

	def __init__(self, inport, outport):
		SimulinkBlock.__init__(self)
		self.inport = [inport]
		self.outport = [outport]
		self.name = "Skip"
		self.skip = True
		
	def isabelle(self, opt_fun, opt_generic, flag):
		return iSkip()


# Main structure for a Simulink block with state variables (e.g., UnitDelay, Integrator, etc.)
class SimulinkStateBlock(SimulinkBlock):

	def add_state_ports(self):
		pi = Port(self.get_id(), 0, "sin", 0)
		po = Port(self.get_id(), 0, "sout", 0)
		self.inport.append(pi)
		self.outport.append(po)

	def update_arities(self):
		flag = 0
		if self.get_inport_arity(0) == 0:
			self.set_inport_arity(0, 1)
		if self.get_inport_arity(0) != self.get_outport_arity(0):
			flag = 1
		self.set_inport_arity(1, self.get_inport_arity(0))
		self.set_outport_arity(0, self.get_inport_arity(0))
		self.set_outport_arity(1, self.get_inport_arity(0))		
		return flag
		
	def set_port_init_val(self, val):
		pi = [p for p in self.inport if p.get_direction() == "sin"][0]
		pi.set_value(str(val))

