# -*- coding: utf-8 -*-

from blocks.simulink_basic import *
from utilities.simulink_utilities import *

# Structure for an Integrator block
class SimulinkIntegrator(SimulinkStateBlock):
	
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkStateBlock.__init__(self, xml, size_in, size_out, parent)
		
		self.set_params("dt")
		
	def split_block(self):
		return [SimulinkBlockIntegratorSingle(self, 0), SimulinkBlockIntegratorSingle(self, 1)]

	def set_init(self):
		val = get_property(self.xml, "InitialCondition")
		self.set_port_init_val(val)

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())

		func_type = "::(real \<times> real => bool) => real \<times> real => bool"
		
		if get_property(self.xml, 'LimitOutput') == 'on':
			high = get_property(self.xml, 'UpperSaturationLimit')
			low = get_property(self.xml, 'LowerSaturationLimit')
			func = iBasic(self.get_type()+"Limit", [high, low] + self.get_params_list())
		else:
			func = iBasic(self.get_type(), self.get_params_list())

		if not opt_generic:
			return "("+func.__repr__()+func_type+")"
		
		return func


# Structure used in spliting Integrator in components with single outputs
class SimulinkBlockIntegratorSingle(SimulinkBlock):
	def __init__(self, integr, select_output):
		SimulinkBlock.__init__(self)
		self.integr = integr
		self.select_output = select_output
		self.outport = [integr.outport[select_output]]
		if select_output == 0:
			self.inport = [integr.inport[1]]
		else:
			self.inport = integr.inport
		self.parent = integr.parent
		self.name = integr.name + ["A", "B"][select_output]
		
	def get_params(self):
		if self.select_output == 0:
			return ""
		else:
			return self.integr.get_params()

	def get_params_list(self):
		if self.select_output == 0:
			return []
		else:
			return self.integr.get_params_list()
			
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			if self.select_output == 0:
				return iBasic(self.get_qualified_name())
			else:
				return iBasic(self.get_qualified_name(), self.get_params_list())
	
		if get_property(self.integr.xml, 'LimitOutput') == 'on':
			params = [get_property(self.integr.xml, 'UpperSaturationLimit'), get_property(self.integr.xml, 'LowerSaturationLimit')] + self.get_params_list()
			limit = "Limit"
		else:
			params = self.get_params_list()
			limit = ""
		
			
		if self.select_output == 0:
			func = iBasic(self.integr.get_type()+ limit + "A")
			func_type = "::(real => bool) => real => bool"

		else:
			func = iBasic(self.integr.get_type() + limit + "B", params)
			func_type = "::(real => bool) => real \<times> real => bool"

		if not opt_generic:
			return "("+func.__repr__()+func_type+")"
		
		return func
		
def get_float(x):
	try:
		return float(x)
	except ValueError as e:
		return None

# Structure for a TransferFcn block
class SimulinkTransferFcn(SimulinkBlock):
	def get_exp_list(self, x):
		a = x.replace("[","").replace("]","").strip()
		b = ""
		while(a != b):
			b = a
			a = a.replace("  ", " ").replace(", ", ",").replace(" ,", ",")
			
		c = a.replace(" ", ",")
		return c.split(",")
		
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)

		self.set_params("dt")
				
		num_aux = self.get_exp_list(get_property(xml, 'Numerator'))
		self.num = []
		for x in num_aux:
			r = get_float(x)
			if r:
				self.num.append(r)
			else:
				self.num.append(x)
		
		
		assert(len(self.num) > 0)
		
		if(get_float(self.num[0]) == 0):
			self.num = self.num[1:]
		
		denom_aux = self.get_exp_list(get_property(xml, 'Denominator'))
		self.denom = []
		for x in denom_aux:
			r = get_float(x)
			if r:
				self.denom.append(r)
			else:
				self.denom.append(x)
						
		assert(len(self.denom) > 0)
		
		if len(self.num) > len(self.denom):
			raise Exception('\nError:\n  The transfer function '+ self.get_qualified_name() +' cannot be translated.\n  The order of the numerator is greater than the order of the denominator.\n')
		
		
	def add_state_ports(self):
		'''
		pi = Port(self.get_id(), 0, "sin", 2)
		self.inport.append(pi)
		po = Port(self.get_id(), 0, "sout", 2)
		self.outport.append(po)
		'''
		
		#handling general case
		if len(self.denom) > 1:
			pi = Port(self.get_id(), 0, "sin", len(self.denom)-1)
			self.inport.append(pi)
			po = Port(self.get_id(), 0, "sout", len(self.denom)-1)
			self.outport.append(po)
		
	def update_arities(self):
		self.outport[0].set_arity(1)
		return 0

	def set_init(self):
		pi = [p for p in self.inport if p.get_direction() == "sin"]
		for p in pi:
			p.set_value(0.0)

	def split_block(self):
		size = len(self.denom)
		out_tuple = self.get_out_tuple()
		return [SimulinkTransferFcnSingle(self, out_tuple[i], i) for i in range(size)]		
		
	def get_out_tuple(self):
		n = len(self.denom) - 1
		
		invars = [p.get_var_name() for p in self.inport]
		inpt = invars[0]
		inpt_st = invars[1:]
		assert(n == len(inpt_st))
		if n:
			s0_aux = " - ".join(["(%s) * %s" % (self.denom[i+1], inpt_st[i]) for i in range(n)])
			s0 = '(%s - %s) / (%s)' % (inpt, s0_aux, self.denom[0])
		else:
			s0 = "%s / (%s)" % (inpt, self.denom[0])
			
		inpt_st_a = [s0] + inpt_st
		
		m = len(self.num)
		
		if m == 0:
			out_tuple = ["0"]
		else:
			out_tuple = [" + ".join(["%s * (%s)" % (inpt_st_a[n +1 - m + i] , self.num[i]) for i in range(m)])]
			
		out_tuple = out_tuple + ["%s + %s * dt" % (inpt_st_a[i+1], inpt_st_a[i]) for i in range(n)]
		
		return out_tuple
		
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())
		
		'''
		func_type = "::(real \<times> real \<times> real => bool) => real \<times> real \<times> real => bool"
		func =  iBasic("TransferTFcn", [self.num[0], self.denom[1]] + self.get_params_list())
		'''
		
		#handling general case

		func_type_aux = " \<times> ".join(['real' for i in range(len(self.inport))])
		func_type = "::(%s \<Rightarrow> bool) \<Rightarrow> %s \<Rightarrow> bool" %(func_type_aux, func_type_aux)
		
		invars = [p.get_var_name() for p in self.inport]
		out_tuple = self.get_out_tuple()
		
		expr = "(" + ",".join(out_tuple) +")"
		
		"""
		#compute output formula
		if len(self.num) == len(self.denom):
			out_expr_1 = '(%s - %s) * (%s / %s) * %s' %(inpt, out_expr_aux, '1.0', self.denom[0], self.num[0])
			out_expr = out_expr_1 + " + " + " + ".join([str(self.num[i+1]) + " * " + inpt_st[i] for i in range(len(inpt_st))])
		else:
			out_expr = " + ".join([str(self.num[i]) + " * " + inpt_st[i] for i in range(len(inpt_st))])
		
		#compute state update formulas	
		out_expr_2 = '(%s - %s) * (%s / %s) * %s' %(inpt, out_expr_aux, '1.0', self.denom[0], 'dt')
		st_expr = [inpt_st[0] + " + " + out_expr_2]
		for i in range(1,len(inpt_st)):
			st_expr.append(inpt_st[i] + " + " + inpt_st[i-1] + " * dt")
		expr = '(%s, %s)' %(out_expr, ",".join(st_expr))
		#expr = [out_expr].append(st_expr)
		"""
		
		if not opt_generic:
			return "(" + str(iFunc(invars, expr, opt_fun)) + func_type + ")"
		
		return iFunc(invars, expr, opt_fun)

# Structure for TransferFcn ... 
class SimulinkTransferFcnSingle(SimulinkBlock):
	def __init__(self, tf, expr, select_output):
		SimulinkBlock.__init__(self)
		
		self.params = tf.get_params_list()
		self.tf = tf
		self.select_output = select_output
		self.outport = [tf.outport[select_output]]

		#dep[0] = self.inport[n + 1 - m:] # +-1 decide if it is OK
		#dep[1] = self.inport
		#dep[2] = self.inport[1] self.inport[2]
		#dep[3] = self.inport[2] self.inport[3]
		# ...
		
		if select_output == 0:
			self.inport = tf.inport[len(tf.denom) - len(tf.num):]
		elif select_output == 1:
			self.inport = tf.inport
		else:
			self.inport = [tf.inport[select_output-1], tf.inport[select_output]]

		self.expr = expr
		
		self.parent = tf.parent
		self.name = tf.name + "_" + self.outport[0].get_var_name()
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())
			
		func_type_aux = " \<times> ".join(['real' for i in range(len(self.inport))])
		func_type = "::(real \<Rightarrow> bool) \<Rightarrow> %s \<Rightarrow> bool" %(func_type_aux)
		
		invars = [p.get_var_name() for p in self.inport]
		
		if not opt_generic:
			return "(" + str(iFunc(invars, self.expr, opt_fun)) + func_type + ")"
		
		return iFunc(invars, self.expr, opt_fun)


