# -*- coding: utf-8 -*-

from blocks.simulink_basic import *
from utilities.simulink_utilities import *

# Structure for a Math block
# types handled: Add, Sum, Subtract, UnaryMinus, Product, Divide, 
class SimulinkMath(SimulinkBlock):
	def update_arities(self):
		# Set by default the arity of the block's outport to 1
		self.set_outport_arity(0,1)
		return 0
	
	# Build the precondition for Product and Divide when divisions are used
	def precondition(self, op_type, ops, vars):
		if op_type == 'Product':
			precTerms = []
			if (ops[0] == "*") or (ops[0] == "/"):
				for i in range(0, len(vars)):
					if ops[i] == '/':
						expr = vars[i]
						for j in range(i + 1, len(vars)):
							expr = expr + ops[j] + vars[j]
						precTerms.append(expr)
			precTerms.reverse()

			if len(precTerms) > 0:
				return ' \<and> '.join(["%s \<noteq> 0" % x for x in precTerms])
		
		return None
	
	# Builds the math expression for the block	
	def expression(self, op_type, ops, vars):
		expr = ''
		if op_type == 'Sum':
			expr = '0'
		else:
			expr = '1'
		#todo: update it to remove 0 or 1.
		#todo: why next test? what else can it be ops[0]? 
		if (op_type == 'Product') and (ops[0] != "*") and (ops[0] != "/"):
			expr = '*'.join(["%s" % (x) for x in vars])
		else:
			for i in range(0, len(vars)):
				expr = expr + ops[i] + vars[i]
		return expr
	
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
			
		# TODO: check case of tuples of inputs per port
		
		# get list of input ports for building formula tuple
		vars = [p.get_var_name() for p in self.inport]
		
		# get type
		op_type = self.get_type()
		
		# if it is a UnaryMinus generate predicate transformer
		if op_type == 'UnaryMinus':
			expr = "-" + vars[0]
			prec = None
		else:
			# get math operators
			ops = get_property(self.xml,'Inputs').replace('|','')
			# build precondition 
			prec = self.precondition(op_type, ops, vars)
			# build math formula based on operators
			expr = self.expression(op_type, ops, vars)
			
		return iFunc(vars, expr, opt_fun, prec)


# Structure for a MathFunction block
class SimulinkMathFunction(SimulinkBlock):
	operators = {'log':'Ln', 'exp':'Expn', 'square':'Square', 'pow':'Power', '10^u': 'Power10'}
	
	def __init__(self,xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self,xml, size_in, size_out, parent)	
	
		self.operator = get_property(xml, 'Operator')
		if self.operator not in SimulinkMathFunction.operators.keys():
			raise Exception('\nError:\n  The operator of the math function ' + self.get_qualified_name() + 'is not handled.\n  The currently available operators are log, exp, square, pow, 10^u.\n')

	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())	
		
		return iBasic(SimulinkMathFunction.operators[self.operator])


# Structure for a Gain block		
class SimulinkGain(SimulinkMath):
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
			
		val = get_property(self.xml,'Gain')  
		for x in fcnMap.keys():
			if val.find(x+"(") !=-1 :
				val = val.replace(x, fcnMap[x])
		for x in cstMap.keys():
			if val.find(x+"(") !=-1 :
				val = val.replace(x, cstMap[x])
		return iBasic("Gain", ["("+val+")"])


# The Simulink Sqrt block is modeled by SimulinkBlock


# Structure for a MinMax block
class SimulinkMinMax(SimulinkBlock):
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
			
		var = [p.get_var_name() for p in self.inport]
		
		op = get_property(self.xml, "Function").upper()
		expr = op + "[" + ",".join(var) + "]"
			
		return iFunc(var, expr, opt_fun, None)



# Structure for a user-defined Function block
class SimulinkFunction(SimulinkBlock):
	def __init__(self,xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self,xml, size_in, size_out, parent)

		# obtain and parse function expression
		expr = get_property(self.xml, 'Expr') 
		parsedExpr = post2tree(inf2post(PP_expr.parseString(expr)))
		
		# function expression
		self.expression = "((%s)::'a::simulink)" % build_expression(parsedExpr)
		self.preconditionDTerms = get_division_preconditions(parsedExpr)
		self.preconditionSRTerms = get_sqrt_preconditions(parsedExpr)
		prec1 = build_precondition(self.preconditionDTerms, "\<noteq> 0")
		prec2 = build_precondition(self.preconditionSRTerms, ">= 0")
		l = []
		if prec1: 
			l.append(prec1)
		if prec2:
			l.append(prec2)
		self.precondition = " \<and> ".join(l)	
		
		# variables on which the function is defined
		self.invariables = get_input_variables(parsedExpr)
		self.multi_arity = True
		
	def update_arities(self):
		# set by default the arity of the block's outport to 1
		self.set_outport_arity(0,1)
		return 0
		
	def split_ports(self):
		SimulinkBlock.split_ports(self)
		for i in range(len(self.inport)):
			self.inport[i].index = i
		self.multi_arity = False

	def split_block(self):
		# functions have one single output port
		# irrelevant inputs are removed such that circular references in feedbacks are avoided
			
		# build the list of input ports not used within the function expression
		self.del_inport = []
        
		inport = self.inport
		self.inport = []
		for i in range(len(inport)):
			if "u%d" % (i + 1) in self.invariables:
				self.inport.append(inport[i])
				inport[i].index = i
			else:
				self.del_inport.append(inport[i])
		return [self]
		
	def get_in_vars(self):
		if self.multi_arity:
			return ['u%s' % (i + 1) for i in range(self.get_inport_arity(0))]
		else:
			return ['u%d' % (p.index + 1) for p in self.inport]
			
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name())
			
		invars = self.get_in_vars()
		# verify that all used variables in the expression have a corresponding input
		for i in self.invariables:
			if i not in invars:
				#error = 'Function %s refers to inexistent variables' % self.expression
				raise Exception('\nError:\n  The function\'s ' + self.get_qualified_name() + ' expression contains inexistent input variables.\n')

		r = append_types(invars, self.invariables)

		return iFunc(invars, self.expression, opt_fun, self.precondition)


# Structure for a SinWave block
class SimulinkSinWave(SimulinkStateBlock):
	def __init__(self, xml, size_in = None, size_out = None, parent = None):
		SimulinkBlock.__init__(self, xml, size_in, size_out, parent)
		
		self.set_params("dt")
		
		self.ampl = get_property(self.xml, 'Amplitude') 
		self.freq = get_property(self.xml, 'Frequency')
		if self.freq.find('pi') != -1 :
			self.freq = self.freq.replace('pi', 's_pi')
		self.phase = get_property(self.xml, 'Phase')
		self.bias = get_property(self.xml, 'Bias')

	def split_block(self):
		return [SimulinkSinWaveSingle(self, 0), SimulinkSinWaveSingle(self, 1)]
		
	def update_arities(self):
		self.set_inport_arity(0, 1)
		self.set_outport_arity(0, 1)
		self.set_outport_arity(1, 1)
		return 0

	def set_init(self):
		self.set_port_init_val(0.0)
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
			return iBasic(self.get_qualified_name(), self.get_params_list())
		return iBasic("SinWaveT", [self.ampl, self.freq, self.phase, self.bias] + self.get_params_list())


# Structure for SinWave variation with one output
class SimulinkSinWaveSingle(SimulinkBlock):
	def __init__(self, sinwave, select_output):
		SimulinkBlock.__init__(self)
		'''
		if select_output == 0:
			self.params = sinwave.get_params_list()
		'''
		self.params = sinwave.get_params_list()
		self.sw = sinwave
		self.select_output = select_output
		self.outport = [sinwave.outport[select_output]]
		self.inport = sinwave.inport
		self.parent = sinwave.parent
		self.name = sinwave.name + ["A", "B"][select_output]
		
	def isabelle(self, opt_fun, opt_generic, flag):
		if not flag:
				return iBasic(self.get_qualified_name(), self.get_params_list())

		if self.select_output == 0:
			return iBasic("SinWaveTA", [self.sw.ampl, self.sw.freq, self.sw.phase, self.sw.bias]+ self.get_params_list())
		else:
			return iBasic("SinWaveTB", self.get_params_list())
		
