# -*- coding: utf-8 -*-

import utilities.simulink_globals 
from utilities.simulink_parser import *
from blocks.simulink_port import *
from copy import copy, deepcopy

# this variable contains the unary functions Simulink defines and their correspondent in Isabelle library
fcnMap = {'sqrt':'s_sqrt', 'exp':'s_exp', 'sin':'s_sin', 'cos':'s_cos', 'abs':'s_abs', 'ln':'s_ln'}
# this variable contains the constants Simulink defines and their correspondent in Isabelle library
cstMap = {'pi':'s_pi'}


# method that builds (recursively) a parsed expression from the corresponding tree    
def build_expression(tree, result = None):
	operators_gen = {"==":"is_eq_num ", "~=":"is_neq_num ", "<":"is_less_num ", "<=":"is_less_eq_num ", ">":"is_gt_num ", ">=":"is_ge_num "}
	#print tree
	if result == None:
		result = ''
	if type(tree) is str:
		return tree

	# unfold the tree as a tuple 
	(operator, operand1, operand2) = tree
	#print (operator, operand1, operand2)
	
	# build expression
	result += '('
	
	
	if type(operand1) is str:
		'''if operand1 == '0':
				pass
		else:
			lhs = operand1
			'''
		lhs = cstMap[operand1] if operand1 in cstMap.keys() else operand1
	# else add the expression to which it corresponds
	else:
		lhs = build_expression(operand1)
		
	if type(operand2) is str :
		rhs = cstMap[operand2] if operand2 in cstMap.keys() else operand2
	else:
		rhs = build_expression(operand2)
		
	if operator in operators_gen.keys():
		res = "%s %s %s" %(operators_gen[operator], lhs, rhs)
	else:
		if operator in fcnMap.keys():
			res = "%s %s" %(fcnMap[operator], lhs)
		else:
			res = "%s %s %s" %(lhs, operator, rhs)
		
	'''
	# if the first operand is a string (except 0) add it to the expression
	if type(operand1) is str:
		if operand1 == '0':
				pass
		#elif operand1 == '1':
		#	result += '('+operand1+'::real)'
		else:
			result += operand1
	# else add the expression to which it corresponds
	else:
		result += build_expression(operand1)

	# add operator
	result += operator
	
	# if the second operand is a string, add it to the expression
	if type(operand2) is str :
		#if operand2 == '1':
		#	result += '('+operand2+'::real)'
		#else:
		result += operand2
	# else add the expression to which it corresponds
	else:
		result += build_expression(operand2)
	
	'''
	result += res
	result += ')'
	
	return result
   

# method that computes the precondition expression of a property transformer based on a list of parsed expressions   
def build_precondition(precList, cond):
	preconditions = []
	
	# for each precondition in the list of parsed expressions
	for p in precList:
		# if it represeted with a list
		if type(p) is list:
			preconditions.append("(%s::'a)" % build_expression(p))
		else:
			preconditions.append("(%s::'a)" % p)
	
	# compute the precondition term (as conjunction)
	pred = " \<and> ".join(["%s %s" % (p, cond) for p in preconditions])

	return pred
	
# method that computes the set of input variables used in a function expression	
def get_input_variables(tree, result = None, level = 0):
	
	if result == None:
		result = []
	
	# 
	if len(tree) == 1:
		return result
	
	# unfold tree
	(operator, operand1, operand2) = tree
	
	# for each operand
	for opnd in [operand1, operand2]:        
		# if it is a 'u' variable append it to the result
		if type(opnd) is str:
			if opnd[0] == 'u':
				#print 'found u!!!'
				result.append(opnd)
		# else explore the next level of the tree
		else:
			get_input_variables(opnd, result, level + 1)
	
	# if all variables have been collected, sort them based on their indices
	if level == 0:
		result=list(set(result))
		if len(result) > 1:
			result.sort(key=lambda x: int(x[1:]))
   
	return result


def append_types(all_invars, used_invars):
	res = [i if i in used_invars else i+"::'a" for i in all_invars]
	return res
	
# method that computes a list of parsed preconditions for an expression 
def get_division_preconditions(tree):

	queue = [tree]
	divisions = []
	
	# for each operator in the tree
	while len(queue) != 0:
		node = queue.pop(0)
		if type(node) is list:
			(operator, operand1, operand2) = node
			if operator in ops:
				queue.append(operand1)
				queue.append(operand2)
				# if it is a division 
				if operator == '/':
					# add the rhs to the list of division preconditions
					divisions.append(operand2)
	
	# sort the preconditions from the shortest (e.g. c1 constant) to the longest 
	depth = lambda L: isinstance(L, list) and max(map(depth, L))+1
	divisions.sort(key=depth)

	return divisions
	
def get_sqrt_preconditions(tree):

	queue = [tree]
	sqrts = []
	
	# for each operator in the tree
	while len(queue) != 0:
		node = queue.pop(0)
		if type(node) is list:
			(operator, operand1, operand2) = node
			if operator in ops:
				queue.append(operand1)
				queue.append(operand2)
				# if it is a division 
				if operator == 'sqrt':
					# add the rhs to the list of division preconditions
					sqrts.append(operand1)
	
	# sort the preconditions from the shortest (e.g. c1 constant) to the longest 
	depth = lambda L: isinstance(L, list) and max(map(depth, L))+1
	sqrts.sort(key=depth)

	return sqrts




class Isabelle:
	def __init__(self, str = None, prec = 0):
		self.str = str
		self.prec = prec

	def is_skip(self):
		return self.str == "Skip"
	
	def __repr__(self):
		return self.str
				
		
class iTuple(Isabelle):
	def __init__(self):
		self.unit = 0
	def mk_pred(self):
		eqs = []
		for var in self.vars.keys():
			eqs.extend(["%s = %s%s" % (var, var, "'" * (i + 1)) for i in range(self.vars[var])])
		return " \<and> ".join(eqs)
		
	def mk_primed_var(self, var):
		if var in self.vars.keys():
			i = self.vars[var] + 1
		else:
			i = 1
		self.vars[var] = i
		return var + "'" * i
	
	def mk_tuple_primed_var(self):
		def fun(vars):
			if vars == []:
				return self.get_unit_var()
			elif len(vars) == 1:
				return self.mk_primed_var(vars[0])
			else:
				return "(" + (", ".join([self.mk_primed_var(v) for v in vars])) + ")"
		return fun
	
	def mk_tuple_var(self):
		return self.mk_tuple(False)
	"""
	def mk_tuple_var_fb(self):
		return self.mk_tuple_fb(False)
		
	def mk_tuple_var_fb_io(self):
		return self.mk_tuple_fb_io(False)
	"""
	def mk_tuple_val(self):
		return self.mk_tuple(True)
	
	def mk_tuple(self, value):
		def fun(vars):
			if vars == []:
				if value:
					return "()"
				else:
					return self.get_unit_var()
			elif len(vars) == 1:
				return vars[0]
			else:
				return "(" + (", ".join(vars)) + ")"
		return fun

	"""
	def mk_tuple_fb_io(self, value):
		def fun(vars):
			if vars == []:
				if value:
					return "()"
				else:
					return self.get_unit_var()
			elif len(vars) == 1:
				#print vars[0]
				return "(" + vars[0] + ", " + self.get_unit_var() + ")"
				'''if len(vars[0]) == 1:
					return "(" + vars[0] + ", " + self.get_unit_var() + "::unit)"
				else:
					return vars[0]
				'''
			else:
				return "(" + (", ".join(vars)) + ")"
		return fun

	def mk_tuple_fb(self, value):
		def fun(vars):
			if vars == []:
				if value:
					return "()"
				else:
					return self.get_unit_var()
			elif len(vars) == 1:
				return "(" + vars[0] + ", " + self.get_unit_var() + ")"
				'''if len(vars[0]) == 1:
					return "(" + vars[0] + ", " + self.get_unit_var() + ")"
				else:
					return vars[0]
				'''
			else:
				iv = ", ".join(vars)
				return "(" + iv + ", " + self.get_unit_var() + ")"
		return fun
	"""	
	def get_unit_var(self):
		self.unit = self.unit + 1
		return "u_%d" % self.unit

# global function which for a list ['a', 'b', 'c'] creates the tuple '(a, b, c)'
mk_tuple_var = iTuple().mk_tuple(False)
mk_tuple_val = iTuple().mk_tuple(True)
#mk_tuple_var_fb = iTuple().mk_tuple_fb(False)
#mk_tuple_var_fb_io = iTuple().mk_tuple_fb_io(False)
	
class iFunc(iTuple):
	def __init__(self, vars, expr, opt_fun, prec = None):
		iTuple.__init__(self)
		tvar = self.mk_tuple_var() (vars)
		
		if prec:
			self.prec = 55
			aprec = "{. %s . %s .} \<circ> " % (tvar, prec)
		else:
			self.prec = 1000
			aprec = ""
		
		if opt_fun:
			self.str = "(%s[- %s \<leadsto> %s -])" % (aprec, tvar, expr)
		else:
			self.str = "(%s[: %s  \<leadsto> res . res = %s :])" % (aprec, tvar, expr)
					
class iSwitchSplit(iTuple):
	# from: [[a,b,c], [x,y,z]], [[x,a], [c,z], [b]]
	# obtain: [:(a,b,c), (x,y,z) -> (x',a'), (c',z'), b' . a' = a /\ ...:]
    
    # example with multiplicities:
	# from: [[a,b^3,c^2], [x,y^2,z]], [[x,a], [c^2,z], [b^3]]
	# obtain: [:(a,b1,b2,b3,c1,c2), (x,y1,y2,z) -> (x',a'), (c1',c2',z'), (b1',b2',b3') . a' = a /\ b1' = b1 ...:]
	
	# The variables on the right hand side may  occur multiple times
	# from: [[a,b^3], [c^2]], [[a,a], [c^2,b^3], [b^3]]
	# obtain: [:(a,b1,b2,b3), (c1,c2) -> (a',a''), (c1',c2',b1',b2'b3'), (b1'',b2'',b3'') . a' = a /\ a'' = a ...:]

	def __init__(self, inpt, outpt, opt_fun):
		self.prec = 1000
		iTuple.__init__(self)
		self.opt_fun = opt_fun
		# map ports with arities to lists of variable names
		flat_in_list_vars  = [map(Port.get_var_list, x) for x in inpt] 
		flat_out_list_vars = [map(Port.get_var_list, x) for x in outpt] 
		f = lambda x: [j for i in x for j in i]
		
		# concatenate the list of variables from the same component 
		self.flat_in_vars   = map(f,  flat_in_list_vars)
		self.flat_out_vars  = map(f,  flat_out_list_vars)
		
		self.flat_in_vars = self.flatten_single(self.flat_in_vars)
		self.flat_out_vars = self.flatten_single(self.flat_out_vars)
		
		#test that all invars are distinct
		all_in = sorted([j for i in self.flat_in_vars for j in i])
		x = None
		for y in all_in:
			assert not x == y
			x = y
		
		
		self.make_switch(opt_fun)
		
	def make_switch(self, opt_fun):
		"""
		if self.check_skip(self.flat_in_vars, self.flat_out_vars):
			self.str = "Skip"		
		elif self.check_skip(self.flat_out_vars, self.flat_in_vars):
			self.str = "Skip"
		"""	
		if self.flat_in_vars == self.flat_out_vars:
			self.str = "Skip"
		elif opt_fun:
			self.mk_update(self.flat_in_vars, self.flat_out_vars)
		else:
			self.mk_demonic(self.flat_in_vars, self.flat_out_vars)

	def flatten_single(self, var_list):
		l = []
		for x in var_list:
			if len(x) == 1:
				l.append(x[0])
			else:
				return var_list
		return [l]

	def check_skip(self, inv, outv):
		is_skip = False
		n = len(inv)
		k = len(outv)
		if not (k == 1 and n == len(outv[0])):
			return False

		for i in range(n):
			if len(inv[i]) == 1 and inv[i][0] == outv[0][i]:
				pass
			else:
				return False
		return True
	
	def mk_update(self, flat_in_vars, flat_out_vars):
		tvar = self.mk_tuple_var() (map(self.mk_tuple_var(), flat_in_vars))	
			
		if tvar[0]=='(':
			tvar = tvar[1:len(tvar)-1]
		
		tval = self.mk_tuple_val() (map(self.mk_tuple_val(), flat_out_vars))
		
		if tval[0]=='(':
			tval = tval[1:len(tval)-1]
		
		self.str = "[- %s \<leadsto> %s -]" % (tvar, tval)

	def mk_demonic(self, flat_in_vars, flat_out_vars):
		self.vars = {}
		tvari = self.mk_tuple_var() (map(self.mk_tuple_var(), flat_in_vars))	
		tvaro = self.mk_tuple_var() (map(self.mk_tuple_primed_var(), flat_out_vars))
		pred = self.mk_pred()
		if pred.strip() == "":
			pred = "True"
		self.str = "[:%s \<leadsto> %s . %s:]" % (tvari, tvaro, pred)


# TODO: if l_i does not match l_o, output error and handle it
class iSwitchCompose(iSwitchSplit):
	def __init__(self, sw1, sw2):
		iTuple.__init__(self)
		self.prec = 1000
		self.opt_fun = sw1.opt_fun
		
		if sw1.is_skip() and sw2.is_skip():
			self.flat_in_vars = sw1.flat_in_vars
			self.flat_out_vars = sw1.flat_out_vars
			self.str = "Skip"
			return
		if sw1.is_skip():
			self.flat_in_vars = sw2.flat_in_vars
			self.flat_out_vars = sw2.flat_out_vars
		elif sw2.is_skip():
			self.flat_in_vars = sw1.flat_in_vars
			self.flat_out_vars = sw1.flat_out_vars
		else:
			self.flat_in_vars = sw1.flat_in_vars
			l_o = sw1.flat_out_vars
			l_i = sw2.flat_in_vars
			#print l_o
			#print l_i
			mapping = {}
			n = len(l_i)
			for i in range(n):
				k = len(l_i[i])
				for j in range(k):
					mapping[l_i[i][j]] = l_o[i][j]
			
			F = lambda x: mapping[x]
			self.flat_out_vars = [map(F, z) for z in sw2.flat_out_vars]
		
			#self.str = 'None'
		self.make_switch(self.opt_fun)

class iSwitch(iSwitchSplit):
	# the same as iSwitchSplit, but requires the same variables for input and output
	def __init__(self, inpt, outpt, opt_fun):
		all_in = [j for i in inpt for j in i]
		all_out = [j for i in outpt for j in i]
		if sorted(all_in) != sorted(all_out) or len(all_in) != len(all_out):
			#print "different variables:", all_in, all_out
			raise Exception('\nError:\n  An internal error has occured.\n  A routing transformer cannot be generated because the input and output variables do not match.\n')
		
		iSwitchSplit.__init__(self, inpt, outpt, opt_fun)
		
# method that
experimental = 1

class iComp(Isabelle):
	def __init__(self, *systems):
		Isabelle.__init__(self)
		
		s = systems
		
		if experimental:
			s = self.flatten_comp(s)
		
		s = [x for x in s if not x.is_skip()]
		if s == []:
			self.str = "Skip"
			self.prec = 1000
			self.systems = []
		else:
			if experimental:
				s = self.colapse_switch(s)
				s = [x for x in s if not x.is_skip()]
			self.systems = s
			if len(s) > 1:
				self.prec = 55
				open_par = {True: "(", False:""}
				close_par = {True: ")", False:""}
				#here we have self.prec > x.prec because \<comp> is associative
				t = [ open_par[self.prec > x.prec] + x.__repr__() + close_par[self.prec > x.prec] for x in s]
				self.str = " \<circ>\n    ".join(t)
			else:
				self.str = s[0].__repr__()
				self.prec = s[0].prec
				
	def flatten_comp(self, systems):
		l = []
		for x in systems:
			if isinstance(x, iComp):
				l = l + x.systems
			else:
				l.append(x)
		return l
		
				
	def colapse_switch(self, systems):
		l = []
		current = None
		for x in systems:
			if isinstance(x, iSwitchSplit):
				if current:
					current = iSwitchCompose(current, x)
					#TODO: test error for iSwitchCompose
					# if current.str == None ....
				else:
					current = x
			else:
				if current:
					l.append(current)
					current = None
				l.append(x)
		if current:
			l.append(current)
		return l
			

class iBasic(Isabelle):
	def __init__(self, name, params = []):
		if params == []:
			Isabelle.__init__(self, name, 1000)
		else:
			n = "%s %s" % (name, "(" + ") (".join(map(str, params))+ ")")
			Isabelle.__init__(self, n, 500)

class iSkip(Isabelle):
	def __init__(self):
		Isabelle.__init__(self, "Skip", 1000)

class iProd(Isabelle):
	def __init__(self, *systems):
		Isabelle.__init__(self)
		k = len(systems) - 1
		#print systems
		while k >= 0 and systems[k].is_skip(): 
			k = k - 1
		
		if k == -1:
			self.str = "Skip"
			self.prec = 1000
		else:			
			if k == 0 and len(systems) == 1:
				self.str = systems[0].__repr__()
				self.prec = systems[0].prec
			else:
				#todo: define these in the Isabelle class as static. remove them from iComp also
				open_par = {True: "(", False:""}
				close_par = {True: ")", False:""}
				self.prec = 70
				#here we have self.prec >= x.prec because ** is NOT associative
				t = [ open_par[self.prec >= x.prec] + x.__repr__() + close_par[self.prec >= x.prec] for x in systems[:k+2]]
				self.str = " ** ".join(t)


class iFeedback(Isabelle):
	def __init__(self, system, n, opt_fun, opt):
		Isabelle.__init__(self)
		self.prec = 500
		'''
		if not opt_fun:
			op = 'feedback'
		else:
			if opt:
				op = 'feedback'
			else:
				op = 'feedbackless'
		'''	
		op = 'feedback'
		
		assert n > 0

		if system.prec <= self.prec:
			fb = "%s (%s)" % (op, system.__repr__())
		else:
			fb = "%s %s" % (op, system.__repr__())			
			
		for i in range(n - 1):
			fb = "%s (%s)" % (op, fb)
		self.str = fb
	
def parse_disj(s):
	s = s.strip()
	vals = []
	if s[:1] == '[':
		s = s[1:]
		v, s = parse_disj(s)
		vals.extend(v)
		while s[:1] == ",":
			s = s[1:]
			v, s = parse_disj(s)
			vals.extend(v)
		s = s.strip()
		if s[:1] != ']':
			raise Exception('\nError:\n  An internal error has occured while parsing the expression '+s+'.\n  The expression must be in the format [ string ].\n') #'expecting ]')
		s = s[1:]
		return vals, s
	else:
		i = 0
		while s[i:i+1] not in [",", "]", ""]:
			i = i + 1
		r = s[:i].strip()
		if r == "":
			return ([], s[i:])
		return ([r], s[i:]) 
		
def parse_conj(s):
	s = s.strip()
	vals = []
	if s[:1] == '{':
		s = s[1:-1]
		v, s = parse_disj(s)
		vals.append(v)
		while s[:1] == ",":
			s = s[1:]
			v, s = parse_disj(s)
			vals.append(v)
		return vals
	else:
		raise Exception('\nError:\n  An internal error has occured while parsing the expression '+s+'.\n  The expression must be in the format { string }.\n')
		
def mk_or(var, rel, exprs):
	return " \<or> ".join(["(%s %s %s)" % (var, rel, e) for e in exprs])

def mk_and(var, rel, exprs):
	return " \<and> ".join(["(%s %s %s)" % (var, rel, e) for e in exprs])


def mk_update(tup, expr):
	return "[- \<lambda> %s . %s -]" %(tup, expr)



def is_head(node, graph):
	flag = True
	for n in graph:
		#print [val for val in b.outport if val in block.inport]
		if (len([val for val in n.outport if val in node.inport]) != 0):
			flag = False
	#print flag
	return flag

def get_dest(edge, graph):
	for node in graph:
		if edge in node.inport:
			return node
			

#method for topological sorting of blocks: not complete, do not use! see method sort_blocks2(li) below
def sort_blocks(li):
	graph = copy.deepcopy(li)
	
	#sorted list
	lo = []
	
	#set of blocks with no "incoming" edges (blocks whose inputs do not belong to the outputs of others)
	s = []
	
	for n in graph:
		aux = [i for i in graph if i != n]
		if is_head(n, aux):
			s.append(n)
	
	#for i in s:
	#	print i.get_id()

	while len(s) != 0:
		n = s.pop()
		src = n
		print "source: ",src.get_type()
		aux = [i for i in graph if i != src]
		
		lo.append([val for val in li if val.get_id() == n.get_id()][0])
		for p in src.outport:
			dest = get_dest(p, aux)
			
			if dest:
				print "dest: ", dest.get_type()
				dest.inport.remove(p)
		
				dest.inport = [p for p in dest.inport if p.get_direction() != "sin"] 
				
				print dest.inport
				
				if len(dest.inport) == 0:
					s.append(dest)
								
		#print "incoming blocks: ", s
		#print "sorting intm: ", lo

	#print "sorted"
	
	for i in lo:
		print i.get_id()
	#	print i.inport
	#	print i.outport
		
	return lo


def get_source(graph):
	source = None
	for n in graph:
		if len(n.inport) == 0:
			source = n
			break
	return source
	
def get_sink(graph):
	sink = None
	for n in graph:
		if len(n.outport) == 0:
			sink = n
			break
	return sink

def get_opt_node(graph, lh, lt, li):
	aux = sorted(graph, key= lambda SimulinkBlock: len(SimulinkBlock.outport) - len(SimulinkBlock.inport), reverse=True)
	#aux2 = [(n, len(n.outport)-len(n.inport)) for n in aux ]
	#aux3 = [n[0] for n in aux2 if n[1] == aux2[0][1]]
	 
	#print aux2
	#print
	#print aux3
	
	#if len(aux3) > 1:
	#	print 'getting node that was connected to a source'
	
	return aux[0]

def inter(la, lb):
	aux = [x for x in la if x in lb]
	return len(aux) > 0

def get_opt_node2(graph, lt, li):
	if len(graph) == 1:
		return graph[0]
	for g in graph:
		elem = get_elem(li, g.get_id())
		aux = [x for x in lt if inter(elem.outport, x.inport)]
		if len(aux) > 0:
			return g

def remove_node(graph, n):
	for px in n.inport:
		src = [v.outport for v in graph if px in v.outport]
		map (lambda s: list.remove(s, px), src)
		#print src
		
	for px in n.outport:
		dest = [v.inport for v in graph if px in v.inport]
		map (lambda d: list.remove(d, px), dest)
		#print dest
		
	#print n
	
	graph.remove(n)

def get_elem(li, sid):
	return [n for n in li if n.get_id() == sid][0]

def sort_blocks2(li):
	graph = deepcopy(li)

	#remove inputs from Inports and outputs from Outports
	for n in graph:
		if n.get_type() == 'Inport':
			n.inport = []
		if n.get_type() == 'Outport':
			n.outport = []
		if n.get_type() == 'Scope':
			n.outport = []
	
	#print
	#print "1. clean inport and outport"
	#print graph
	
	#print
	#print "2. feedback arc set"
	
	lh = []
	lt = []
	
	while len(graph) != 0:
		#print "iter"
		source = get_source(graph)
		#print "source ", source
		sink = get_sink(graph)
		#print "sink ", sink
		n = None
		
		if sink is not None:
			#print 'sink: ', sink
			lt.insert(0,get_elem(li, sink.get_id()))
			n = sink
		elif source is not None:
			#print 'source: ',source
			lh.append(get_elem(li, source.get_id()))
			n = source
		else:
			#n = get_opt_node(graph, lh, lt, li)
			n = get_opt_node2(graph, lt, li)
			#print 'opt: ', n
			#lh.append(get_elem(li, n.get_id()))
			lt.insert(0,get_elem(li, n.get_id()))
		remove_node(graph, n)
	
	res = lh + lt
	#print res
	return res
	

